
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "xvariables.h"
#include "initglxctx.h"
#include "xwidgets.h"
#include "meshes.h"
#include "palm.h"
#include "lights.h"
#include "app3_a.h"
#include "app3_b.h"

#define MESHDEG 3

GPUmesh *mymesh[NPALMMESHES];
float   model_rot_axis[3] = {0.0,1.0,0.0},
        model_rot_angle0 = 0.0, model_rot_angle;
float   viewer_pos0[4] = {0.0,0.0,10.0,1.0}, viewer_pos[4];

GLfloat edges_colour[4] = {0.0,0.5,0.7,1.0};
GLfloat facets_colour[4] = {0.91,0.65,0.5};

void LoadMyShaders ( void )
{
  LoadMeshRefinementShaders ();
  LoadMeshRenderingShaders ();
} /*LoadMyShaders*/

void DestroyMyShaders ( void )
{
  DestroyMeshRefinementShaders ();
  DestroyMeshRenderingShaders ();
} /*DestroyMyShaders*/

void InitPalmMeshes ( void )
{
  int i;

  if ( (mymesh[0] = EnterPalmToGPU ()) ) {
    for ( i = 1; i < NPALMMESHES; i++ ) {
      if ( !(mymesh[i] = malloc ( sizeof(GPUmesh) )) )
        ExitOnError ( "InitPalmMeshes 0" );
      memset ( mymesh[i], 0, sizeof(GPUmesh) );
      if ( !GPUmeshRefinement ( MESHDEG, mymesh[i-1], mymesh[i] ) )
        ExitOnError ( "InitPalmMeshes 1" );
      printf ( "%d: nv = %d, nhe = %d, nfac = %d\n",
               i, mymesh[i]->nv, mymesh[i]->nhe, mymesh[i]->nfac );
    }
  }
  else
    ExitOnError ( "InitPalmMeshes" );
} /*InitPalmMeshes*/

void InitLights ( void )
{
  GLfloat amb0[4] = { 0.2, 0.2, 0.3, 1.0 };
  GLfloat dif0[4] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetLightAmbient ( 0, amb0 );
  SetLightDiffuse ( 0, dif0 );
  SetLightPosition ( 0, pos0 );
  SetLightAttenuation ( 0, atn0 );
  SetLightOnOff ( 0, 1 );
} /*InitLights*/

void InitMyObject ( void )
{
  TimerInit ();
  memset ( &myglwdata, 0, sizeof(myglwindata) );
  myglwdata.speed = 0.5*3.1415926;
  memset ( &light, 0, sizeof(LightBl) );
  SetupModelMatrix ( &myglwdata, model_rot_axis, model_rot_angle );
  InitViewMatrix ( &myglwdata );
  InitLights ();
  InitPalmMeshes ();
  ExitIfGLError ( "InitMyObject" );
} /*InitMyObject*/

void SetVPMatrix ( GLfloat vm[16], GLfloat pm[16], GLfloat ep[4] )
{
  GLfloat vpm[16];

  M4x4Multf ( vpm, pm, vm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 4*sizeof(GLfloat), ep );
  ExitIfGLError ( "SetVPMatrix" );
} /*SetVPMatrix*/

void SetupModelMatrix ( myglwindata *wd, float axis[3], float angle )
{
  M4x4RotateVf ( wd->mm, axis[0], axis[1], axis[2], angle );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), wd->mm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), wd->mm );
  ExitIfGLError ( "SetupModelMatrix" );
} /*SetupModelMatrix*/

void InitViewMatrix ( myglwindata *wd )
{
  memcpy ( wd->eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( wd->vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), wd->vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 4*sizeof(GLfloat), wd->eyepos );
  ExitIfGLError ( "InitViewMatrix" );
} /*InitViewMatrix*/

void SetupViewMatrix ( myglwindata *wd, float vk[3], float angk )
{
  GLfloat tm[16];

  memcpy ( wd->viewer_rvec, vk, 3*sizeof(float) );
  wd->viewer_rangle = angk;
  M4x4Translatef ( wd->vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4MRotateVf ( wd->vm, wd->viewer_rvec[0], wd->viewer_rvec[1], wd->viewer_rvec[2],
                  -wd->viewer_rangle );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  M4x4Transposef ( tm, wd->vm );  tm[3] = tm[7] = tm[11] = 0.0;
  M4x4MultMVf ( wd->eyepos, tm, viewer_pos0 );
  ExitIfGLError ( "SetupViewMatrix" );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
} /*SetupViewMatrix*/

char RotateViewer ( myglwindata *wd, float width, float height,
                    int delta_xi, int delta_eta )
{
  float   vi[3], lgt, angi, vk[3], angk;

  if ( delta_xi == 0 && delta_eta == 0 )
    return false;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(wd->rl)/height;
  vi[1] = (float)delta_xi*(wd->tb)/width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, wd->viewer_rvec,
                     wd->viewer_rangle, vi, angi );
  SetupViewMatrix ( wd, vk, angk );
  return true;
} /*RotateViewer*/

void DrawMyScene ( void )
{
  int i;

  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable ( GL_DEPTH_TEST );
  if ( meshsw[0] )
    DrawMeshEdges ( mymesh[0], edges_colour );
  for ( i = 1; i < NPALMMESHES; i++ )
    if ( meshsw[i] )
      DrawMeshFacets ( mymesh[i], facets_colour );
} /*DrawMyScene*/

void DestroyMyObject ( void )
{
  int i;

  for ( i = 0; i < NPALMMESHES; i++ )
    DeleteGPUmesh ( mymesh[i] );
  ExitIfGLError ( "DestroyMyObject" );
} /*DestroyMyObject*/

