
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "xvariables.h"
#include "initglxctx.h"
#include "xwidgets.h"
#include "bezpatches.h"
#include "meshes.h"
#include "linkage.h"
#include "palm.h"
#include "lights.h"
#include "app3c_a.h"
#include "app3c_b.h"

#define MESHDEG 3

float   model_rot_axis[3] = {0.0,1.0,0.0};
float   viewer_pos0[4] = {0.0,0.0,10.0,1.0}, viewer_pos[4];

void LoadMyShaders ( void )
{
  LoadLinkageArticulationShaders ();
  LoadMeshRefinementShaders ();
  LoadMeshNormalVectorShaders ();
  LoadRenderingShaders ();
} /*LoadMyShaders*/

void DestroyMyShaders ( void )
{
  DestroyLinkageArticulationShaders ();
  DestroyMeshRefinementShaders ();
  DestroyMeshNormalVectorShaders ();
  DestroyRenderingShaders ();
} /*DestroyMyShaders*/

void InitLights ( void )
{
  GLfloat amb0[4] = { 0.2, 0.2, 0.3, 1.0 };
  GLfloat dif0[4] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };
  GLfloat csc[3]  = { 0.0, 0.0, 0.0 };

  SetLightAmbient ( 0, amb0 );
  SetLightDiffuse ( 0, dif0 );
  SetLightPosition ( 0, pos0 );
  SetLightAttenuation ( 0, atn0 );
  SetLightOnOff ( 0, 1 );
  ConstructShadowTxtFBO ( 0 );
  SetupShadowTxtTransformations ( 0, csc, 2.2 );
  UpdateShadowMatrix ( 0 );
} /*InitLights*/

void InitMyObject ( void )
{
  TimerInit ();
  memset ( &myglwdata, 0, sizeof(myglwindata) );
  myglwdata.speed = 0.5*3.1415926;
  memset ( &light, 0, sizeof(LightBl) );
  SetupModelMatrix ( &myglwdata, model_rot_axis, 0.0 );
  InitViewMatrix ( &myglwdata );
  InitLights ();
  memset ( &myscene, 0, sizeof(mypalmscene) );
  myscene.meshsw[0] = myscene.meshsw[2] = true;
  myscene.lod = /*-1*/ 2;
  ConstructPalmLinkage ( &myscene );
  ArticulatePalmLinkage ( &myscene );
  ExitIfGLError ( "InitMyObject" );
} /*InitMyObject*/

void SetVPMatrix ( GLfloat vm[16], GLfloat pm[16], GLfloat ep[4] )
{
  GLfloat vpm[16];

  M4x4Multf ( vpm, pm, vm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), ep );
  ExitIfGLError ( "SetVPMatrix" );
} /*SetVPMatrix*/

void SetupModelMatrix ( myglwindata *wd, float axis[3], float angle )
{
  wd->model_rot_angle = angle;
  M4x4RotateVf ( wd->mm, axis[0], axis[1], axis[2], angle );
  memcpy ( wd->mmti, wd->mm, 16*sizeof(GLfloat) );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), wd->mm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), wd->mmti );
  ExitIfGLError ( "SetupModelMatrix" );
} /*SetupModelMatrix*/

void InitViewMatrix ( myglwindata *wd )
{
  memcpy ( wd->eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( wd->vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), wd->vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), wd->eyepos );
  ExitIfGLError ( "InitViewMatrix" );
} /*InitViewMatrix*/

void SetupViewMatrix ( myglwindata *wd, float vk[3], float angk )
{
  GLfloat tm[16];

  memcpy ( wd->viewer_rvec, vk, 3*sizeof(float) );
  wd->viewer_rangle = angk;
  M4x4Translatef ( wd->vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4MRotateVf ( wd->vm, wd->viewer_rvec[0], wd->viewer_rvec[1], wd->viewer_rvec[2],
                  -wd->viewer_rangle );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  M4x4Transposef ( tm, wd->vm );  tm[3] = tm[7] = tm[11] = 0.0;
  M4x4MultMVf ( wd->eyepos, tm, viewer_pos0 );
  ExitIfGLError ( "SetupViewMatrix" );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
} /*SetupViewMatrix*/

char RotateViewer ( myglwindata *wd, float width, float height,
                    int delta_xi, int delta_eta )
{
  float   vi[3], lgt, angi, vk[3], angk;

  if ( delta_xi == 0 && delta_eta == 0 )
    return false;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(wd->rl)/height;
  vi[1] = (float)delta_xi*(wd->tb)/width;
  vi[2] = 0.0;   
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, wd->viewer_rvec,
                     wd->viewer_rangle, vi, angi );
  SetupViewMatrix ( wd, vk, angk );
  return true;
} /*RotateViewer*/

void DrawScene ( mypalmscene *scene )
{
  kl_Redraw ( scene->lkg );
} /*DrawScene*/

void DrawSceneToShadows ( mypalmscene *scene )
{
  int    l;
  GLuint mask;

  scene->draw_shadows = true;
  glViewport ( 0, 0, SHADOW_MAP_SIZE, SHADOW_MAP_SIZE );
  glEnable ( GL_POLYGON_OFFSET_FILL );
  glPolygonOffset ( 2.0f, 4.0f );
  for ( l = 0, mask = 0x00000001;  l < light.nls;  l++, mask <<= 1 )
    if ( light.shmask & mask ) {
      BindShadowTxtFBO ( l );
      glClear ( GL_DEPTH_BUFFER_BIT );
      DrawScene ( scene );
    }
  glBindFramebuffer ( GL_FRAMEBUFFER, 0 );
  glDisable ( GL_POLYGON_OFFSET_FILL );
  for ( l = 0, mask = 0x00000001;  l < light.nls;  l++, mask <<= 1 )
    if ( light.shmask & mask ) {
      glActiveTexture ( GL_TEXTURE0+l );
      glBindTexture ( GL_TEXTURE_2D, light.ls[l].shadow_txt );
    }
  scene->draw_shadows = false;
} /*DrawSceneToShadows*/

void DrawSceneToWindow ( mypalmscene *scene, myglwindata *wd, int w, int h )
{
/*
  glEnable ( GL_SAMPLE_SHADING );
  glMinSampleShading ( 1.0 );
*/
  glEnable ( GL_DEPTH_TEST );
  DrawSceneToShadows ( scene );
  glViewport ( 0, 0, w, h );
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
  DrawScene ( scene );
} /*DrawSceneToWindow*/

void DestroyMyObject ( void )
{
  kl_DestroyLinkage ( myscene.lkg );
  ExitIfGLError ( "DestroyMyObject" );
} /*DestroyMyObject*/

